import json
import logging
import uuid
from http import HTTPStatus
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

from fastapi import HTTPException, Response
from fastapi.responses import FileResponse
from starlette.requests import Request
from watchfiles import Change, awatch
from jupyverse_api.auth import Auth, User
from jupyverse_api.kernels import Kernels, KernelsConfig
from jupyverse_api.kernels.models import CreateSession, Execution, Kernel, Notebook, Session
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.yjs import Yjs
from jupyverse_api.app import App

from .kernel_driver.driver import KernelDriver
from .kernel_driver.kernelspec import find_kernelspec, kernelspec_dirs
from .kernel_server.server import (
    AcceptedWebSocket,
    KernelServer,
    kernels,
)


logger = logging.getLogger("kernels")


class _Kernels(Kernels):
    def __init__(
        self,
        app: App,
        kernels_config: KernelsConfig,
        auth: Auth,
        frontend_config: FrontendConfig,
        yjs: Optional[Yjs],
    ) -> None:
        super().__init__(app=app, auth=auth)
        self.kernels_config = kernels_config
        self.frontend_config = frontend_config
        self.yjs = yjs

        self.kernelspecs: dict = {}
        self.kernel_id_to_connection_file: Dict[str, str] = {}
        self.sessions: Dict[str, Session] = {}
        self.kernels = kernels

    async def get_kernelspecs(
        self,
        user: User,
    ):
        for search_path in kernelspec_dirs():
            for path in Path(search_path).glob("*/kernel.json"):
                with open(path) as f:
                    spec = json.load(f)
                name = path.parent.name
                resources = {
                    f.stem: f"{self.frontend_config.base_url}kernelspecs/{name}/{f.name}"
                    for f in path.parent.iterdir()
                    if f.is_file() and f.name != "kernel.json"
                }
                self.kernelspecs[name] = {"name": name, "spec": spec, "resources": resources}
        return {"default": self.kernels_config.default_kernel, "kernelspecs": self.kernelspecs}

    async def get_kernelspec(
        self,
        kernel_name,
        file_name,
        user: User,
    ):
        for search_path in kernelspec_dirs():
            file_path = Path(search_path) / kernel_name / file_name
            if file_path.exists():
                return FileResponse(file_path)

        raise HTTPException(
            status_code=404, detail=f"Kernelspec {kernel_name}/{file_name} not found"
        )

    async def get_kernels(
        self,
        user: User,
    ):
        results = []
        for kernel_id, kernel in kernels.items():
            if kernel["server"]:
                connections = kernel["server"].connections
                last_activity = kernel["server"].last_activity["date"]
                execution_state = kernel["server"].last_activity["execution_state"]
            else:
                connections = 0
                last_activity = ""
                execution_state = "idle"
            results.append(
                {
                    "id": kernel_id,
                    "name": kernel["name"],
                    "connections": connections,
                    "last_activity": last_activity,
                    "execution_state": execution_state,
                }
            )
        return results

    async def delete_session(
        self,
        session_id: str,
        user: User,
    ):
        kernel_id = self.sessions[session_id].kernel.id
        kernel_server = kernels[kernel_id]["server"]
        await kernel_server.stop()
        del kernels[kernel_id]
        if kernel_id in self.kernel_id_to_connection_file:
            del self.kernel_id_to_connection_file[kernel_id]
        del self.sessions[session_id]
        return Response(status_code=HTTPStatus.NO_CONTENT.value)

    async def rename_session(
        self,
        request: Request,
        user: User,
    ):
        rename_session = await request.json()
        session_id = rename_session.pop("id")
        for key, value in rename_session.items():
            setattr(self.sessions[session_id], key, value)
        return self.sessions[session_id]

    async def get_sessions(
        self,
        user: User,
    ):
        for session in self.sessions.values():
            kernel_id = session.kernel.id
            if kernel_id in kernels:
                kernel_server = kernels[kernel_id]["server"]
                session.kernel.last_activity = kernel_server.last_activity["date"]
                session.kernel.execution_state = kernel_server.last_activity["execution_state"]
        return list(self.sessions.values())

    async def create_session(
        self,
        request: Request,
        user: User,
    ):
        create_session = CreateSession(**(await request.json()))
        kernel_id = create_session.kernel.id
        kernel_name = create_session.kernel.name
        if kernel_name is not None:
            # launch a new ("internal") kernel
            kernel_cwd = Path(create_session.path).parent
            while True:
                if kernel_cwd.is_dir():
                    break
                kernel_cwd = kernel_cwd.parent
            kernel_server = KernelServer(
                kernelspec_path=Path(find_kernelspec(kernel_name)).as_posix(),
                kernel_cwd=str(kernel_cwd),
            )
            kernel_id = str(uuid.uuid4())
            kernels[kernel_id] = {"name": kernel_name, "server": kernel_server, "driver": None}
            await kernel_server.start()
        elif kernel_id is not None:
            # external kernel
            kernel_name = kernels[kernel_id]["name"]
            kernel_server = KernelServer(
                connection_file=self.kernel_id_to_connection_file[kernel_id],
                write_connection_file=False,
            )
            kernels[kernel_id]["server"] = kernel_server
            await kernel_server.start(launch_kernel=False)
        else:
            return
        session_id = str(uuid.uuid4())
        session = Session(
            id=session_id,
            path=create_session.path,
            name=create_session.name,
            type=create_session.type,
            kernel=Kernel(
                id=kernel_id,
                name=kernel_name,
                connections=kernel_server.connections,
                last_activity=kernel_server.last_activity["date"],
                execution_state=kernel_server.last_activity["execution_state"],
            ),
            notebook=Notebook(
                path=create_session.path,
                name=create_session.name,
            ),
        )
        self.sessions[session_id] = session
        return session

    async def interrupt_kernel(
        self,
        kernel_id,
        user: User,
    ):
        if kernel_id in kernels:
            kernel = kernels[kernel_id]
            kernel["server"].interrupt()
            result = {
                "id": kernel_id,
                "name": kernel["name"],
                "connections": kernel["server"].connections,
                "last_activity": kernel["server"].last_activity["date"],
                "execution_state": kernel["server"].last_activity["execution_state"],
            }
            return result

    async def restart_kernel(
        self,
        kernel_id,
        user: User,
    ):
        if kernel_id in kernels:
            kernel = kernels[kernel_id]
            await kernel["server"].restart()
            result = {
                "id": kernel_id,
                "name": kernel["name"],
                "connections": kernel["server"].connections,
                "last_activity": kernel["server"].last_activity["date"],
                "execution_state": kernel["server"].last_activity["execution_state"],
            }
            return result

    async def execute_cell(
        self,
        request: Request,
        kernel_id,
        user: User,
    ):
        if self.yjs is None:
            raise RuntimeError("Cannot execute without a Yjs plugin.")

        r = await request.json()
        execution = Execution(**r)
        if kernel_id in kernels:
            ynotebook = self.yjs.get_document(execution.document_id)
            cell = ynotebook.get_cell(execution.cell_idx)
            cell["outputs"] = []

            kernel = kernels[kernel_id]
            if not kernel["driver"]:
                kernel["driver"] = driver = KernelDriver(
                    kernelspec_path=Path(find_kernelspec(kernel["name"])).as_posix(),
                    write_connection_file=False,
                    connection_file=kernel["server"].connection_file_path,
                )
                await driver.connect()
            driver = kernel["driver"]

            await driver.execute(cell)
            ynotebook.set_cell(execution.cell_idx, cell)

    async def get_kernel(
        self,
        kernel_id,
        user: User,
    ):
        if kernel_id in kernels:
            kernel = kernels[kernel_id]
            result = {
                "id": kernel_id,
                "name": kernel["name"],
                "connections": kernel["server"].connections,
                "last_activity": kernel["server"].last_activity["date"],
                "execution_state": kernel["server"].last_activity["execution_state"],
            }
            return result

    async def shutdown_kernel(
        self,
        kernel_id,
        user: User,
    ):
        if kernel_id in kernels:
            await kernels[kernel_id]["server"].stop()
            del kernels[kernel_id]
        for session_id in [k for k, v in self.sessions.items() if v.kernel.id == kernel_id]:
            del self.sessions[session_id]
        return Response(status_code=HTTPStatus.NO_CONTENT.value)

    async def kernel_channels(
        self,
        kernel_id,
        session_id,
        websocket_permissions,
    ):
        if websocket_permissions is None:
            return
        websocket, permissions = websocket_permissions
        subprotocol = (
            "v1.kernel.websocket.jupyter.org"
            if "v1.kernel.websocket.jupyter.org" in websocket["subprotocols"]
            else None
        )
        await websocket.accept(subprotocol=subprotocol)
        accepted_websocket = AcceptedWebSocket(websocket, subprotocol)
        if kernel_id in kernels:
            kernel_server = kernels[kernel_id]["server"]
            if kernel_server is None:
                # this is an external kernel
                # kernel is already launched, just start a kernel server
                kernel_server = KernelServer(
                    connection_file=self.kernel_id_to_connection_file[kernel_id],
                    write_connection_file=False,
                )
                await kernel_server.start(launch_kernel=False)
                kernels[kernel_id]["server"] = kernel_server
            await kernel_server.serve(accepted_websocket, session_id, permissions)

    async def watch_connection_files(self, path: Path) -> None:
        # first time scan, treat everything as added files
        initial_changes = {(Change.added, str(p)) for p in path.iterdir()}
        await self.process_connection_files(initial_changes)
        # then, on every change
        async for changes in awatch(path):
            await self.process_connection_files(changes)

    async def process_connection_files(self, changes: Set[Tuple[Change, str]]):
        # get rid of "simultaneously" added/deleted files
        file_changes: Dict[str, List[Change]] = {}
        for c in changes:
            change, path = c
            if path not in file_changes:
                file_changes[path] = []
            file_changes[path].append(change)
        to_delete: List[str] = []
        for p, cs in file_changes.items():
            if Change.added in cs and Change.deleted in cs:
                cs.remove(Change.added)
                cs.remove(Change.deleted)
                if not cs:
                    to_delete.append(p)
        for p in to_delete:
            del file_changes[p]
        # process file changes
        for path, cs in file_changes.items():
            for change in cs:
                if change == Change.deleted:
                    if path in self.kernel_id_to_connection_file.values():
                        kernel_id = list(self.kernel_id_to_connection_file.keys())[
                            list(self.kernel_id_to_connection_file.values()).index(path)
                        ]
                        del self.kernel_id_to_connection_file[kernel_id]
                        del kernels[kernel_id]
                elif change == Change.added:
                    try:
                        data = json.loads(Path(path).read_text())
                    except Exception:
                        continue
                    if "kernel_name" not in data or "key" not in data:
                        continue
                    # looks like a kernel connection file
                    kernel_id = str(uuid.uuid4())
                    self.kernel_id_to_connection_file[kernel_id] = path
                    kernels[kernel_id] = {
                        "name": data["kernel_name"],
                        "server": None,
                        "driver": None,
                    }



import httpx
from asphalt.core import Component, ContainerComponent, Context, context_teardown
from jupyverse_api.auth import Auth, AuthConfig
from jupyverse_api.app import App
from sqlalchemy.ext.asyncio import AsyncEngine, AsyncSession

from .config import AuthJupyterHubConfig
from .db import Base
from .routes import auth_factory


class _AuthJupyterHubComponent(Component):
    @context_teardown
    async def start(
        self,
        ctx: Context,
    ) -> None:
        app = await ctx.request_resource(App)
        db_session = await ctx.request_resource(AsyncSession)
        db_engine = await ctx.request_resource(AsyncEngine)

        http_client = httpx.AsyncClient()
        auth_jupyterhub = auth_factory(app, db_session, http_client)
        ctx.add_resource(auth_jupyterhub, types=Auth)

        async with db_engine.begin() as conn:
            await conn.run_sync(Base.metadata.create_all)

        yield

        await http_client.aclose()


class AuthJupyterHubComponent(ContainerComponent):
    def __init__(self, **kwargs):
        self.auth_jupyterhub_config = AuthJupyterHubConfig(**kwargs)
        super().__init__()

    async def start(
        self,
        ctx: Context,
    ) -> None:
        ctx.add_resource(self.auth_jupyterhub_config, types=AuthConfig)
        self.add_component(
            "sqlalchemy",
            url=self.auth_jupyterhub_config.db_url,
        )
        self.add_component("auth_jupyterhub", type=_AuthJupyterHubComponent)
        await super().start(ctx)



from asphalt.core import Component, Context
from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.contents import Contents

from .routes import _Contents


class ContentsComponent(Component):
    async def start(
        self,
        ctx: Context,
    ) -> None:
        app = await ctx.request_resource(App)
        auth = await ctx.request_resource(Auth)  # type: ignore

        contents = _Contents(app, auth)
        ctx.add_resource(contents, types=Contents)



import json
import logging
import os
from glob import glob
from http import HTTPStatus
from pathlib import Path
from typing import List, Optional, Tuple

import json5  # type: ignore
import pkg_resources
from babel import Locale
from fastapi import Response, status
from fastapi.responses import FileResponse, RedirectResponse
from starlette.requests import Request

from jupyverse_api.app import App
from jupyverse_api.auth import Auth, User
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.jupyterlab import JupyterLabConfig
from jupyverse_api.lab import Lab


logger = logging.getLogger("lab")


class _Lab(Lab):
    def __init__(
        self,
        app: App,
        auth: Auth,
        frontend_config: FrontendConfig,
        jupyterlab_config: Optional[JupyterLabConfig] = None,
    ) -> None:
        super().__init__(app, auth, jupyterlab_config)

        self.frontend_config = frontend_config
        self.locale = "en"

    async def get_root(
        self,
        response: Response,
        user: User,
    ):
        # auto redirect
        response.status_code = status.HTTP_302_FOUND
        response.headers["Location"] = self.frontend_config.base_url + self.redirect_after_root

    async def get_favicon(self):
        return FileResponse(Path(__file__).parent / "static" / "favicon.ico")

    async def get_mathjax(self, rest_of_path):
        return RedirectResponse(
            "https://cdnjs.cloudflare.com/ajax/libs/mathjax/2.7.7/" + rest_of_path
        )

    async def get_listings(self, user: User):
        return {
            "blocked_extensions_uris": [],
            "allowed_extensions_uris": [],
            "blocked_extensions": [],
            "allowed_extensions": [],
        }

    async def get_extensions(self, user: User):
        return self.federated_extensions

    async def get_translations_(
        self,
        user: User,
    ):
        return RedirectResponse(f"{self.frontend_config.base_url}lab/api/translations")

    async def get_translations(self, user: User):
        locale = Locale.parse("en")
        display_name = (locale.get_display_name(self.locale) or "").capitalize()
        native_name = (locale.get_display_name() or "").capitalize()
        data = {
            "en": {
                "displayName": display_name,
                "nativeName": native_name,
            }
        }
        for ep in pkg_resources.iter_entry_points(group="jupyterlab.languagepack"):
            locale = Locale.parse(ep.name)
            data[ep.name] = {
                "displayName": display_name,
                "nativeName": native_name,
            }
        return {"data": data, "message": ""}

    async def get_translation(
        self,
        language,
        user: User,
    ):
        if language == "en":
            self.locale = language
            return {}

        for ep in pkg_resources.iter_entry_points(group="jupyterlab.languagepack"):
            if ep.name == language:
                break
        else:
            return {"data": {}, "message": f"Language pack '{language}' not installed!"}
        self.locale = language
        package = ep.load()
        data = {}
        for path in (Path(package.__file__).parent / "locale" / language / "LC_MESSAGES").glob(
            "*.json"
        ):
            with open(path) as f:
                data.update({path.stem: json.load(f)})
        return {"data": data, "message": ""}

    async def get_setting(
        self,
        name0,
        name1,
        name2,
        user: User,
    ):
        with open(self.jlab_dir / "static" / "package.json") as f:
            package = json.load(f)
        if name0 in ["@jupyterlab", "@retrolab"]:
            schemas_parent = self.jlab_dir
        else:
            schemas_parent = self.extensions_dir / name0 / name1
        with open(schemas_parent / "schemas" / name0 / name1 / f"{name2}.json") as f:
            schema = json.load(f)
        key = f"{name1}:{name2}"
        setting = {
            "id": f"@jupyterlab/{key}",
            "schema": schema,
            "version": package["version"],
            "raw": "{}",
            "settings": {},
            "last_modified": None,
            "created": None,
        }
        if user:
            user_settings = json.loads(user.settings)
            if key in user_settings:
                setting.update(user_settings[key])
                setting["settings"] = json5.loads(user_settings[key]["raw"])
        return setting

    async def change_setting(
        self,
        request: Request,
        name0,
        name1,
        user: User,
        user_update,
    ):
        settings = json.loads(user.settings)
        settings[f"{name0}:{name1}"] = await request.json()
        await user_update({"settings": json.dumps(settings)})
        return Response(status_code=HTTPStatus.NO_CONTENT.value)

    async def get_settings(self, user: User):
        if user:
            user_settings = json.loads(user.settings)
        else:
            user_settings = {}
        settings = []
        schemas = [self.jlab_dir / "schemas"]
        for d1 in self.labextensions_dir.iterdir():
            if (d1 / "schemas").exists():
                schemas.append(d1 / "schemas")
            for d2 in d1.iterdir():
                if (d2 / "schemas").exists():
                    schemas.append(d2 / "schemas")
        for s in schemas:
            for d1 in s.iterdir():
                for d2 in d1.iterdir():
                    package = json.loads((d2 / "package.json.orig").read_text())
                    for path in [p for p in d2.iterdir() if p.suffix == ".json"]:
                        schema = json.loads(path.read_text())
                        key = f"{path.parent.name}:{path.stem}"
                        setting = {
                            "id": f"{d1.name}/{key}",
                            "schema": schema,
                            "version": package["version"],
                            "raw": "{}",
                            "settings": {},
                            "warning": None,
                            "last_modified": None,
                            "created": None,
                        }
                        if key in user_settings:
                            setting.update(user_settings[key])
                            setting["settings"] = json5.loads(user_settings[key]["raw"])
                        settings.append(setting)
        return {"settings": settings}

    def get_federated_extensions(self, extensions_dir: Path) -> Tuple[List, List]:
        federated_extensions = []
        disabled_extensions = []

        for path in glob(os.path.join(extensions_dir, "**", "package.json"), recursive=True):
            with open(path) as f:
                package = json.load(f)
            if "jupyterlab" not in package:
                continue
            extension = package["jupyterlab"]["_build"]
            extension["name"] = package["name"]
            extension["description"] = package["description"]
            federated_extensions.append(extension)

            for ext in package["jupyterlab"].get("disabledExtensions", []):
                disabled_extensions.append(ext)

        return federated_extensions, disabled_extensions



import base64
import json
import os
import shutil
from datetime import datetime
from http import HTTPStatus
from pathlib import Path
from typing import Dict, List, Optional, Union, cast

from anyio import open_file
from fastapi import HTTPException, Response
from jupyverse_api.auth import User
from jupyverse_api.contents import Contents
from jupyverse_api.contents.models import (
    Checkpoint,
    Content,
    CreateContent,
    RenameContent,
    SaveContent,
)
from starlette.requests import Request

from .fileid import FileIdManager


class _Contents(Contents):
    async def create_checkpoint(
        self,
        path,
        user: User,
    ):
        src_path = Path(path)
        dst_path = Path(".ipynb_checkpoints") / f"{src_path.stem}-checkpoint{src_path.suffix}"
        try:
            dst_path.parent.mkdir(exist_ok=True)
            shutil.copyfile(src_path, dst_path)
        except Exception:
            # FIXME: return error code?
            return []
        mtime = get_file_modification_time(dst_path)
        return Checkpoint(**{"id": "checkpoint", "last_modified": mtime})

    async def create_content(
        self,
        path: Optional[str],
        request: Request,
        user: User,
    ):
        create_content = CreateContent(**(await request.json()))
        content_path = Path(create_content.path)
        if create_content.type == "notebook":
            available_path = get_available_path(content_path / "Untitled.ipynb")
            async with await open_file(available_path, "w") as f:
                await f.write(
                    json.dumps({"cells": [], "metadata": {}, "nbformat": 4, "nbformat_minor": 5})
                )
            src_path = available_path
            dst_path = Path(".ipynb_checkpoints") / f"{src_path.stem}-checkpoint{src_path.suffix}"
            try:
                dst_path.parent.mkdir(exist_ok=True)
                shutil.copyfile(src_path, dst_path)
            except Exception:
                # FIXME: return error code?
                pass
        elif create_content.type == "directory":
            name = "Untitled Folder"
            available_path = get_available_path(content_path / name, sep=" ")
            available_path.mkdir(parents=True, exist_ok=True)
        else:
            assert create_content.ext is not None
            available_path = get_available_path(content_path / ("untitled" + create_content.ext))
            open(available_path, "w").close()

        return await self.read_content(available_path, False)

    async def get_root_content(
        self,
        content: int,
        user: User,
    ):
        return await self.read_content("", bool(content))

    async def get_checkpoint(
        self,
        path,
        user: User,
    ):
        src_path = Path(path)
        dst_path = Path(".ipynb_checkpoints") / f"{src_path.stem}-checkpoint{src_path.suffix}"
        if not dst_path.exists():
            return []
        mtime = get_file_modification_time(dst_path)
        return [Checkpoint(**{"id": "checkpoint", "last_modified": mtime})]

    async def get_content(
        self,
        path: str,
        content: int,
        user: User,
    ):
        return await self.read_content(path, bool(content))

    async def save_content(
        self,
        path,
        request: Request,
        response: Response,
        user: User,
    ):
        content = SaveContent(**(await request.json()))
        try:
            await self.write_content(content)
        except Exception:
            raise HTTPException(status_code=404, detail=f"Error saving {content.path}")
        return await self.read_content(content.path, False)

    async def delete_content(
        self,
        path,
        user: User,
    ):
        p = Path(path)
        if p.exists():
            if p.is_dir():
                shutil.rmtree(p)
            else:
                p.unlink()
        return Response(status_code=HTTPStatus.NO_CONTENT.value)

    async def rename_content(
        self,
        path,
        request: Request,
        user: User,
    ):
        rename_content = RenameContent(**(await request.json()))
        Path(path).rename(rename_content.path)
        return await self.read_content(rename_content.path, False)

    async def read_content(
        self, path: Union[str, Path], get_content: bool, file_format: Optional[str] = None
    ) -> Content:
        if isinstance(path, str):
            path = Path(path)
        content: Optional[Union[str, Dict, List[Dict]]] = None
        if get_content:
            if path.is_dir():
                content = [
                    (await self.read_content(subpath, get_content=False)).model_dump()
                    for subpath in path.iterdir()
                    if not subpath.name.startswith(".")
                ]
            elif path.is_file() or path.is_symlink():
                try:
                    async with await open_file(path, mode="rb") as f:
                        content_bytes = await f.read()
                    if file_format == "base64":
                        content = base64.b64encode(content_bytes).decode("ascii")
                    elif file_format == "json":
                        content = json.loads(content_bytes)
                    else:
                        content = content_bytes.decode()
                except Exception:
                    raise HTTPException(status_code=404, detail="Item not found")
        format: Optional[str] = None
        if path.is_dir():
            size = None
            type = "directory"
            format = "json"
            mimetype = None
        elif path.is_file() or path.is_symlink():
            size = get_file_size(path)
            if path.suffix == ".ipynb":
                type = "notebook"
                format = None
                mimetype = None
                if content is not None:
                    nb: dict
                    if file_format == "json":
                        content = cast(Dict, content)
                        nb = content
                    else:
                        content = cast(str, content)
                        nb = json.loads(content)
                    for cell in nb["cells"]:
                        if "metadata" not in cell:
                            cell["metadata"] = {}
                        cell["metadata"].update({"trusted": False})
                    if file_format != "json":
                        content = json.dumps(nb)
            elif path.suffix == ".json":
                type = "json"
                format = "text"
                mimetype = "application/json"
            else:
                type = "file"
                format = None
                mimetype = "text/plain"
        else:
            raise HTTPException(status_code=404, detail="Item not found")

        return Content(
            **{
                "name": path.name,
                "path": path.as_posix(),
                "last_modified": get_file_modification_time(path),
                "created": get_file_creation_time(path),
                "content": content,
                "format": format,
                "mimetype": mimetype,
                "size": size,
                "writable": is_file_writable(path),
                "type": type,
            }
        )

    async def write_content(self, content: Union[SaveContent, Dict]) -> None:
        if not isinstance(content, SaveContent):
            content = SaveContent(**content)
        if content.format == "base64":
            async with await open_file(content.path, "wb") as f:
                content.content = cast(str, content.content)
                content_bytes = content.content.encode("ascii")
                await f.write(content_bytes)
        else:
            async with await open_file(content.path, "wt") as f:
                if content.format == "json":
                    dict_content = cast(Dict, content.content)
                    if content.type == "notebook":
                        # see https://github.com/jupyterlab/jupyterlab/issues/11005
                        if (
                            "metadata" in dict_content
                            and "orig_nbformat" in dict_content["metadata"]
                        ):
                            del dict_content["metadata"]["orig_nbformat"]
                    await f.write(json.dumps(dict_content, indent=2))
                else:
                    content.content = cast(str, content.content)
                    await f.write(content.content)

    @property
    def file_id_manager(self):
        return FileIdManager()


def get_available_path(path: Path, sep: str = "") -> Path:
    directory = path.parent
    name = Path(path.name)
    i = None
    while True:
        if i is None:
            i_str = ""
            i = 1
        else:
            i_str = str(i)
            i += 1
        if i_str:
            i_str = sep + i_str
        available_path = directory / (name.stem + i_str + name.suffix)
        if not available_path.exists():
            return available_path


def get_file_modification_time(path: Path):
    if path.exists():
        return datetime.utcfromtimestamp(path.stat().st_mtime).isoformat() + "Z"


def get_file_creation_time(path: Path):
    if path.exists():
        return datetime.utcfromtimestamp(path.stat().st_ctime).isoformat() + "Z"


def get_file_size(path: Path) -> Optional[int]:
    if path.exists():
        return path.stat().st_size
    raise HTTPException(status_code=404, detail="Item not found")


def is_file_writable(path: Path) -> bool:
    if path.exists():
        if path.is_dir():
            # FIXME
            return True
        else:
            return os.access(path, os.W_OK)
    return False



from abc import ABC, abstractmethod

from fastapi import APIRouter, Depends
from fastapi.responses import HTMLResponse
from jupyverse_api import Router

from ..app import App
from ..auth import Auth, User
from ..lab import Lab


class RetroLab(Router, ABC):
    def __init__(self, app: App, auth: Auth, lab: Lab):
        super().__init__(app=app)

        router = APIRouter()

        @router.get("/retro/tree", response_class=HTMLResponse)
        async def get_tree(
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_tree(user)

        @router.get("/retro/notebooks/{path:path}", response_class=HTMLResponse)
        async def get_notebook(
            path,
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_notebook(path, user)

        @router.get("/retro/edit/{path:path}", response_class=HTMLResponse)
        async def edit_file(
            path,
            user: User = Depends(auth.current_user()),
        ):
            return await self.edit_file(path, user)

        @router.get("/retro/consoles/{path:path}", response_class=HTMLResponse)
        async def get_console(
            path,
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_console(path, user)

        @router.get("/retro/terminals/{name}", response_class=HTMLResponse)
        async def get_terminal(
            name: str,
            user: User = Depends(auth.current_user()),
        ):
            return await self.get_terminal(name, user)

        self.include_router(router)

    @abstractmethod
    async def get_tree(
        self,
        user: User,
    ):
        ...

    @abstractmethod
    async def get_notebook(
        self,
        path,
        user: User,
    ):
        ...

    @abstractmethod
    async def edit_file(
        self,
        path,
        user: User,
    ):
        ...

    @abstractmethod
    async def get_console(
        self,
        path,
        user: User,
    ):
        ...

    @abstractmethod
    async def get_terminal(
        self,
        name: str,
        user: User,
    ):
        ...



from typing import Optional

from pydantic import BaseModel


class KernelInfo(BaseModel):
    name: Optional[str] = None
    id: Optional[str] = None


class CreateSession(BaseModel):
    kernel: KernelInfo
    name: str
    path: str
    type: str


class Kernel(BaseModel):
    id: str
    name: str
    last_activity: str
    execution_state: str
    connections: int


class Notebook(BaseModel):
    path: str
    name: str


class Session(BaseModel):
    id: str
    path: str
    name: str
    type: str
    kernel: Kernel
    notebook: Notebook


class Execution(BaseModel):
    document_id: str
    cell_idx: int



import os
import sys

from .paths import jupyter_data_dir

if os.name == "nt":
    programdata = os.environ.get("PROGRAMDATA", None)
    if programdata:
        SYSTEM_JUPYTER_PATH = [os.path.join(programdata, "jupyter")]
    else:  # PROGRAMDATA is not defined by default on XP
        SYSTEM_JUPYTER_PATH = [os.path.join(sys.prefix, "share", "jupyter")]
else:
    SYSTEM_JUPYTER_PATH = [
        "/usr/local/share/jupyter",
        "/usr/share/jupyter",
    ]

ENV_JUPYTER_PATH = [os.path.join(sys.prefix, "share", "jupyter")]


def jupyter_path(*subdirs):
    paths = []
    # highest priority is env
    if os.environ.get("JUPYTER_PATH"):
        paths.extend(p.rstrip(os.sep) for p in os.environ["JUPYTER_PATH"].split(os.pathsep))
    # then user dir
    paths.append(jupyter_data_dir())
    # then sys.prefix
    for p in ENV_JUPYTER_PATH:
        if p not in SYSTEM_JUPYTER_PATH:
            paths.append(p)
    # finally, system
    paths.extend(SYSTEM_JUPYTER_PATH)

    # add subdir, if requested
    if subdirs:
        paths = [os.path.join(p, *subdirs) for p in paths]
    return paths


def kernelspec_dirs():
    return jupyter_path("kernels")


def _is_kernel_dir(path):
    return os.path.isdir(path) and os.path.isfile(os.path.join(path, "kernel.json"))


def _list_kernels_in(kernel_dir):
    if kernel_dir is None or not os.path.isdir(kernel_dir):
        return {}
    kernels = {}
    for f in os.listdir(kernel_dir):
        path = os.path.join(kernel_dir, f)
        if _is_kernel_dir(path):
            key = f.lower()
            kernels[key] = path
    return kernels


def find_kernelspec(kernel_name):
    d = {}
    for kernel_dir in kernelspec_dirs():
        kernels = _list_kernels_in(kernel_dir)
        for kname, spec in kernels.items():
            if kname not in d:
                d[kname] = os.path.join(spec, "kernel.json")
    return d.get(kernel_name, "")



from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Tuple

from fastapi import Depends, HTTPException, Request, Response, WebSocket, status
from fastapi.security import APIKeyCookie
from fief_client import FiefAccessTokenInfo, FiefAsync, FiefUserInfo
from fief_client.integrations.fastapi import FiefAuth
from jupyverse_api.auth import User

from .config import _AuthFiefConfig


@dataclass
class Res:
    fief: FiefAsync
    session_cookie_name: str
    auth: FiefAuth
    current_user: Any
    update_user: Any
    websocket_auth: Any


def get_backend(auth_fief_config: _AuthFiefConfig) -> Res:
    class CustomFiefAuth(FiefAuth):
        client: FiefAsync

        async def get_unauthorized_response(self, request: Request, response: Response):
            if auth_fief_config.callback_url:
                redirect_uri = auth_fief_config.callback_url
            else:
                redirect_uri = str(request.url_for("auth_callback"))
            auth_url = await self.client.auth_url(redirect_uri, scope=["openid", "offline_access"])
            raise HTTPException(
                status_code=status.HTTP_307_TEMPORARY_REDIRECT,
                headers={"Location": auth_url},
            )

    fief = FiefAsync(
        auth_fief_config.base_url,
        auth_fief_config.client_id,
        auth_fief_config.client_secret,
    )

    session_cookie_name = "fps_auth_fief_user_session"
    scheme = APIKeyCookie(name=session_cookie_name, auto_error=False)
    auth = CustomFiefAuth(fief, scheme)

    async def update_user(
        user: FiefUserInfo = Depends(auth.current_user()),
        access_token_info: FiefAccessTokenInfo = Depends(auth.authenticated()),
    ):
        async def _(data: Dict[str, Any]) -> FiefUserInfo:
            user = await fief.update_profile(access_token_info["access_token"], {"fields": data})
            return user

        return _

    def websocket_auth(permissions: Optional[Dict[str, List[str]]] = None):
        async def _(
            websocket: WebSocket,
        ) -> Optional[Tuple[WebSocket, Optional[Dict[str, List[str]]]]]:
            accept_websocket = False
            checked_permissions: Optional[Dict[str, List[str]]] = None
            if session_cookie_name in websocket._cookies:
                access_token = websocket._cookies[session_cookie_name]
                if permissions is None:
                    accept_websocket = True
                else:
                    checked_permissions = {}
                    for resource, actions in permissions.items():
                        allowed = checked_permissions[resource] = []
                        for action in actions:
                            try:
                                await fief.validate_access_token(
                                    access_token, required_permissions=[f"{resource}:{action}"]
                                )
                            except BaseException:
                                pass
                            else:
                                allowed.append(action)
                                accept_websocket = True
            if accept_websocket:
                return websocket, checked_permissions
            else:
                await websocket.close(code=status.WS_1008_POLICY_VIOLATION)
                return None

        return _

    def current_user(permissions=None):
        if permissions is not None:
            permissions = [
                f"{resource}:{action}"
                for resource, actions in permissions.items()
                for action in actions
            ]

        async def _(
            user: FiefUserInfo = Depends(auth.current_user(permissions=permissions)),
        ):
            return User(**user["fields"])

        return _

    return Res(
        fief=fief,
        session_cookie_name=session_cookie_name,
        auth=auth,
        current_user=current_user,
        update_user=update_user,
        websocket_auth=websocket_auth,
    )



from asphalt.core import Component, Context
from jupyverse_api.auth import AuthConfig
from jupyverse_api.login import Login
from jupyverse_api.app import App

from .routes import _Login


class LoginComponent(Component):
    async def start(
        self,
        ctx: Context,
    ) -> None:
        app = await ctx.request_resource(App)
        auth_config = await ctx.request_resource(AuthConfig)

        login = _Login(app, auth_config)
        ctx.add_resource(login, types=Login)



import glob
import os
import sys
import tempfile
import uuid
from typing import Dict, List


def _expand_path(s):
    if os.name == "nt":
        i = str(uuid.uuid4())
        s = s.replace("$\\", i)
    s = os.path.expandvars(os.path.expanduser(s))
    if os.name == "nt":
        s = s.replace(i, "$\\")
    return s


def _filefind(filename, path_dirs=()):
    filename = filename.strip('"').strip("'")
    if os.path.isabs(filename) and os.path.isfile(filename):
        return filename

    path_dirs = path_dirs or ("",)

    for path in path_dirs:
        if path == ".":
            path = os.getcwd()
        testname = _expand_path(os.path.join(path, filename))
        if os.path.isfile(testname):
            return os.path.abspath(testname)

    raise IOError(f"File {filename} does not exist in any of the search paths: {path_dirs}")


def get_home_dir():
    home = os.path.expanduser("~")
    home = os.path.realpath(home)
    return home


_dtemps: Dict = {}


def _mkdtemp_once(name):
    if name in _dtemps:
        return _dtemps[name]
    d = _dtemps[name] = tempfile.mkdtemp(prefix=name + "-")
    return d


def jupyter_config_dir():
    if os.environ.get("JUPYTER_NO_CONFIG"):
        return _mkdtemp_once("jupyter-clean-cfg")
    if "JUPYTER_CONFIG_DIR" in os.environ:
        return os.environ.env["JUPYTER_CONFIG_DIR"]
    home = get_home_dir()
    return os.path.join(home, ".jupyter")


def jupyter_data_dir():
    if "JUPYTER_DATA_DIR" in os.environ:
        return os.environ["JUPYTER_DATA_DIR"]

    home = get_home_dir()

    if sys.platform == "darwin":
        return os.path.join(home, "Library", "Jupyter")
    elif os.name == "nt":
        appdata = os.environ.get("APPDATA", None)
        if appdata:
            return os.path.join(appdata, "jupyter")
        else:
            return os.path.join(jupyter_config_dir(), "data")
    else:
        xdg = os.environ.get("XDG_DATA_HOME", None)
        if not xdg:
            xdg = os.path.join(home, ".local", "share")
        return os.path.join(xdg, "jupyter")


def jupyter_runtime_dir():
    if "JUPYTER_RUNTIME_DIR" in os.environ:
        return os.environ("JUPYTER_RUNTIME_DIR")
    return os.path.join(jupyter_data_dir(), "runtime")


def find_connection_file(
    filename: str = "kernel-*.json",
    paths: List[str] = [],
) -> str:
    if not paths:
        paths = [".", jupyter_runtime_dir()]

    path = _filefind(filename, paths)
    if path:
        return path

    if "*" in filename:
        pat = filename
    else:
        pat = f"*{filename}*"

    matches = []
    for p in paths:
        matches.extend(glob.glob(os.path.join(p, pat)))

    matches = [os.path.abspath(m) for m in matches]
    if not matches:
        raise IOError(f"Could not find {filename} in {paths}")
    elif len(matches) == 1:
        return matches[0]
    else:
        return sorted(matches, key=lambda f: os.stat(f).st_atime)[-1]



from __future__ import annotations

import logging
from collections import defaultdict
from typing import Dict, List

from fastapi import FastAPI

from ..exceptions import RedirectException, _redirect_exception_handler


logger = logging.getLogger("app")


class App:
    """A wrapper around FastAPI that checks for endpoint path conflicts."""

    _app: FastAPI
    _router_paths: Dict[str, List[str]]

    def __init__(self, app: FastAPI, mount_path: str | None = None):
        if mount_path is None:
            self._app = app
        else:
            subapi = FastAPI()
            app.mount(mount_path, subapi)
            self._app = subapi
        app.add_exception_handler(RedirectException, _redirect_exception_handler)
        self._router_paths = defaultdict(list)

    @property
    def _paths(self):
        return [path for router, paths in self._router_paths.items() for path in paths]

    def _include_router(self, router, _type, **kwargs) -> None:
        new_paths = []
        for route in router.routes:
            path = kwargs.get("prefix", "") + route.path
            for _router, _paths in self._router_paths.items():
                if path in _paths:
                    raise RuntimeError(
                        f"{_type} adds a handler for a path that is already defined in "
                        f"{_router}: {path}"
                    )
            logger.debug("%s added handler for path: %s", _type, path)
            new_paths.append(path)
        self._router_paths[_type].extend(new_paths)
        self._app.include_router(router, **kwargs)

    def _mount(self, path: str, _type, *args, **kwargs) -> None:
        for _router, _paths in self._router_paths.items():
            if path in _paths:
                raise RuntimeError(
                    f"{_type } mounts a path that is already defined in {_router}: {path}"
                )
        self._router_paths[_type].append(path)
        logger.debug("%s mounted path: %s", _type, path)
        self._app.mount(path, *args, **kwargs)

    def add_middleware(self, middleware, *args, **kwargs) -> None:
        self._app.add_middleware(middleware, *args, **kwargs)



from asphalt.core import Component, Context
from jupyverse_api.app import App
from jupyverse_api.auth import Auth
from jupyverse_api.frontend import FrontendConfig
from jupyverse_api.jupyterlab import JupyterLab, JupyterLabConfig
from jupyverse_api.lab import Lab

from .routes import _JupyterLab


class JupyterLabComponent(Component):
    def __init__(self, **kwargs):
        self.jupyterlab_config = JupyterLabConfig(**kwargs)

    async def start(
        self,
        ctx: Context,
    ) -> None:
        ctx.add_resource(self.jupyterlab_config, types=JupyterLabConfig)

        app = await ctx.request_resource(App)
        auth = await ctx.request_resource(Auth)  # type: ignore
        frontend_config = await ctx.request_resource(FrontendConfig)
        lab = await ctx.request_resource(Lab)  # type: ignore

        jupyterlab = _JupyterLab(app, self.jupyterlab_config, auth, frontend_config, lab)
        ctx.add_resource(jupyterlab, types=JupyterLab)



import asyncio
import os
import time
import uuid
from typing import Any, Dict, List, Optional, cast

from .connect import cfg_t, connect_channel, launch_kernel, read_connection_file
from .connect import write_connection_file as _write_connection_file
from .kernelspec import find_kernelspec
from .message import create_message, receive_message, send_message


def deadline_to_timeout(deadline: float) -> float:
    return max(0, deadline - time.time())


class KernelDriver:
    def __init__(
        self,
        kernel_name: str = "",
        kernelspec_path: str = "",
        kernel_cwd: str = "",
        connection_file: str = "",
        write_connection_file: bool = True,
        capture_kernel_output: bool = True,
    ) -> None:
        self.capture_kernel_output = capture_kernel_output
        self.kernelspec_path = kernelspec_path or find_kernelspec(kernel_name)
        self.kernel_cwd = kernel_cwd
        if not self.kernelspec_path:
            raise RuntimeError("Could not find a kernel, maybe you forgot to install one?")
        if write_connection_file:
            self.connection_file_path, self.connection_cfg = _write_connection_file(connection_file)
        else:
            self.connection_file_path = connection_file
            self.connection_cfg = read_connection_file(connection_file)
        self.key = cast(str, self.connection_cfg["key"])
        self.session_id = uuid.uuid4().hex
        self.msg_cnt = 0
        self.execute_requests: Dict[str, Dict[str, asyncio.Future]] = {}
        self.channel_tasks: List[asyncio.Task] = []

    async def restart(self, startup_timeout: float = float("inf")) -> None:
        for task in self.channel_tasks:
            task.cancel()
        msg = create_message("shutdown_request", content={"restart": True})
        await send_message(msg, self.control_channel, self.key, change_date_to_str=True)
        while True:
            msg = cast(
                Dict[str, Any], await receive_message(self.control_channel, change_str_to_date=True)
            )
            if msg["msg_type"] == "shutdown_reply" and msg["content"]["restart"]:
                break
        await self._wait_for_ready(startup_timeout)
        self.channel_tasks = []
        self.listen_channels()

    async def start(self, startup_timeout: float = float("inf"), connect: bool = True) -> None:
        self.kernel_process = await launch_kernel(
            self.kernelspec_path,
            self.connection_file_path,
            self.kernel_cwd,
            self.capture_kernel_output,
        )
        if connect:
            await self.connect(startup_timeout)

    async def connect(self, startup_timeout: float = float("inf")) -> None:
        self.connect_channels()
        await self._wait_for_ready(startup_timeout)
        self.listen_channels()

    def connect_channels(self, connection_cfg: Optional[cfg_t] = None):
        connection_cfg = connection_cfg or self.connection_cfg
        self.shell_channel = connect_channel("shell", connection_cfg)
        self.control_channel = connect_channel("control", connection_cfg)
        self.iopub_channel = connect_channel("iopub", connection_cfg)

    def listen_channels(self):
        self.channel_tasks.append(asyncio.create_task(self.listen_iopub()))
        self.channel_tasks.append(asyncio.create_task(self.listen_shell()))

    async def stop(self) -> None:
        self.kernel_process.kill()
        await self.kernel_process.wait()
        os.remove(self.connection_file_path)
        for task in self.channel_tasks:
            task.cancel()

    async def listen_iopub(self):
        while True:
            msg = await receive_message(self.iopub_channel, change_str_to_date=True)
            msg_id = msg["parent_header"].get("msg_id")
            if msg_id in self.execute_requests.keys():
                self.execute_requests[msg_id]["iopub_msg"].set_result(msg)

    async def listen_shell(self):
        while True:
            msg = await receive_message(self.shell_channel, change_str_to_date=True)
            msg_id = msg["parent_header"].get("msg_id")
            if msg_id in self.execute_requests.keys():
                self.execute_requests[msg_id]["shell_msg"].set_result(msg)

    async def execute(
        self,
        cell: Dict[str, Any],
        timeout: float = float("inf"),
        msg_id: str = "",
        wait_for_executed: bool = True,
    ) -> None:
        if cell["cell_type"] != "code":
            return
        content = {"code": cell["source"], "silent": False}
        msg = create_message(
            "execute_request", content, session_id=self.session_id, msg_id=str(self.msg_cnt)
        )
        if msg_id:
            msg["header"]["msg_id"] = msg_id
        else:
            msg_id = msg["header"]["msg_id"]
        self.msg_cnt += 1
        await send_message(msg, self.shell_channel, self.key, change_date_to_str=True)
        if wait_for_executed:
            deadline = time.time() + timeout
            self.execute_requests[msg_id] = {
                "iopub_msg": asyncio.Future(),
                "shell_msg": asyncio.Future(),
            }
            while True:
                try:
                    await asyncio.wait_for(
                        self.execute_requests[msg_id]["iopub_msg"],
                        deadline_to_timeout(deadline),
                    )
                except asyncio.TimeoutError:
                    error_message = f"Kernel didn't respond in {timeout} seconds"
                    raise RuntimeError(error_message)
                msg = self.execute_requests[msg_id]["iopub_msg"].result()
                self._handle_outputs(cell["outputs"], msg)
                if (
                    msg["header"]["msg_type"] == "status"
                    and msg["content"]["execution_state"] == "idle"
                ):
                    break
                self.execute_requests[msg_id]["iopub_msg"] = asyncio.Future()
            try:
                await asyncio.wait_for(
                    self.execute_requests[msg_id]["shell_msg"],
                    deadline_to_timeout(deadline),
                )
            except asyncio.TimeoutError:
                error_message = f"Kernel didn't respond in {timeout} seconds"
                raise RuntimeError(error_message)
            msg = self.execute_requests[msg_id]["shell_msg"].result()
            cell["execution_count"] = msg["content"]["execution_count"]
            del self.execute_requests[msg_id]

    async def _wait_for_ready(self, timeout):
        deadline = time.time() + timeout
        new_timeout = timeout
        while True:
            msg = create_message(
                "kernel_info_request", session_id=self.session_id, msg_id=str(self.msg_cnt)
            )
            self.msg_cnt += 1
            await send_message(msg, self.shell_channel, self.key, change_date_to_str=True)
            msg = await receive_message(
                self.shell_channel, timeout=new_timeout, change_str_to_date=True
            )
            if msg is None:
                error_message = f"Kernel didn't respond in {timeout} seconds"
                raise RuntimeError(error_message)
            if msg["msg_type"] == "kernel_info_reply":
                msg = await receive_message(
                    self.iopub_channel, timeout=0.2, change_str_to_date=True
                )
                if msg is not None:
                    break
            new_timeout = deadline_to_timeout(deadline)

    def _handle_outputs(self, outputs: List[Dict[str, Any]], msg: Dict[str, Any]):
        msg_type = msg["header"]["msg_type"]
        content = msg["content"]
        if msg_type == "stream":
            if (not outputs) or (outputs[-1]["name"] != content["name"]):
                outputs.append({"name": content["name"], "output_type": msg_type, "text": []})
            outputs[-1]["text"].append(content["text"])
        elif msg_type in ("display_data", "execute_result"):
            outputs.append(
                {
                    "data": {"text/plain": [content["data"].get("text/plain", "")]},
                    "execution_count": content["execution_count"],
                    "metadata": {},
                    "output_type": msg_type,
                }
            )
        elif msg_type == "error":
            outputs.append(
                {
                    "ename": content["ename"],
                    "evalue": content["evalue"],
                    "output_type": "error",
                    "traceback": content["traceback"],
                }
            )
        else:
            return
